from google.cloud import storage

def get_type(filename):
    filetype = filename.split('.')[1]
    if filetype == 'json':
        return 'application/json'
    if filetype == 'txt':
        return 'text/plain'
    if filetype=='csv':
        return 'text/csv'

# 	retirado de: https://cloud.google.com/storage/docs/uploading-objects#storage-upload-object-code-sample
def upload_blob_from_memory(bucket_name, contents, destination_blob_name):
    storage_client = storage.Client\
        .from_service_account_json('service_account.json')
    bucket = storage_client.bucket(bucket_name)
    blob = bucket.blob(destination_blob_name)

    content_type = get_type(destination_blob_name)
    blob.upload_from_string(contents, content_type=content_type)



import pandas
from datetime import datetime

class virtual_machine_table_comprehensive:
    def replace_ip_id_with_ip_val(id_list, pipdf):
        ip_vals = []
        for i in id_list:
            row = pipdf.loc[pipdf['public_ip_id']==i]
            if not row.empty:
                ip_val = row['public_ip_address'].values[0]
            else:
                ip_val = 'public_ip_id not found'
            ip_vals.append(ip_val)
        return ip_vals

    def vm_network_table(data):
        # Creates an auxiliary table by joining public and private
        # IPs, using network card ID as primary key.
        # Useful for retrieving IP data for each VM through their
        # network cards.
        private_vnw = [
            {
                'networkCard_id': id,
                'private_ip': data['network']['private'][id]
            } for id in data['network']['private'].keys()
        ]
        private_vnw_df = pandas.json_normalize(private_vnw)
        public_nw = [
            {
                'networkCard_id': id,
                'public_ip_ids': data['network']['public'][id]
            } for id in data['network']['public'].keys()
        ]

        for item in public_nw:
            item['public_ip_ids'] = [
                ip_resource['id'] for ip_resource in item['public_ip_ids']
            ]

        public_nw_df = pandas.json_normalize(public_nw)
        public_ip_addr = [
            {
                'public_ip_id': id,
                'public_ip_address': data['public_ip_address'][id]
            } for id in data['public_ip_address'].keys()
        ]
        public_ip_df = pandas.json_normalize(public_ip_addr)
        ids = public_nw_df['public_ip_ids']
        ips = ids.apply(virtual_machine_table_comprehensive\
            .replace_ip_id_with_ip_val, args=(public_ip_df,))
        ips = ips.rename('public_ips')
        public_nw_df = pandas.concat([public_nw_df, ips], axis=1)
        vm_network_df = pandas.merge(private_vnw_df, public_nw_df, 
                                 how='outer',
                                 on='networkCard_id')
        return vm_network_df


    def instance_type_info(name, it_df):
        name = name.upper()
        return it_df[it_df['name']\
                .apply(lambda x: x.upper()) == name.upper()]

    def query_addresses(net_cards, nwdf, public=False):
        addresses = []
        if public:
            for net_card in net_cards:
                ips = nwdf[nwdf['networkCard_id'] == net_card['id']]\
                    ['public_ips'].values[0]
                if type(ips) == list:
                    addresses += ips
        else:    
            for net_card in net_cards:
                ips = nwdf[nwdf['networkCard_id'] == net_card['id']]\
                    ['private_ip'].values[0]
                addresses += ips
        return ';'.join(addresses)

    def create_table(data):
        # ----- auxiliary data
        vm_network_df = virtual_machine_table_comprehensive.vm_network_table(data)
        instance_type_df = pandas.json_normalize(data['instance_type'])
        vms = pandas.json_normalize(data['virtual_machine'])
        date = datetime.today().strftime('%Y-%m-%d')

        tags = vms['id'].apply(
                lambda x: 
                    data['resource_group'][x.split('/')[4].upper()])

        subscription_id_name = vms['id'].apply(
                lambda x: 
                    data['subscription'].get(x.split('/')[2]))

        project_id = tags.apply(
                lambda x: 
                    x['ProjectName'] if 'ProjectName' in x.keys()
                    else '')

        it_df = pandas.json_normalize(data['instance_type'])
        it_info = vms['properties.hardwareProfile.vmSize']\
                .apply(virtual_machine_table_comprehensive.instance_type_info,
                       args=(it_df,))
        it_info = pandas.concat(it_info.values, axis=0)
        # -----

        # Data to be saved in tables. Each key in the DataFrame roughly
        # corresponds to a column in the database.
        vmTable = pandas.DataFrame({ 
            'vm_msid': vms['properties.vmId'],

            'networkcard_count': vms['properties.networkProfile'\
                '.networkInterfaces'].apply(lambda x: len(x)),

            'private_ip_address': vms['properties.networkProfile.networkInterfaces']\
                .apply(virtual_machine_table_comprehensive.query_addresses, args=(vm_network_df, False,)),

            'public_ip_address': vms['properties.networkProfile.networkInterfaces']\
                .apply(virtual_machine_table_comprehensive.query_addresses, args=(vm_network_df, True,)),

            'computer_name': vms['name'],

            'resource_name': vms['properties.osProfile.computerName'],
            
            'operating_system_id': vms['properties.storageProfile'\
                '.osDisk.osType'],

            'subscription_id.name': subscription_id_name,

            'subscription_id.subscription_msid': vms['id']\
                .apply(lambda x: x.split('/')[2]),

            'image_id.name': vms['properties.storageProfile'\
                '.imageReference.offer'],

            'image_id.publisher': vms['properties.storageProfile'\
                '.imageReference.publisher'],

            'image_id.sku': vms['properties.storageProfile'\
                '.imageReference.sku'],

            'image_id.version': vms['properties.storageProfile'\
                '.imageReference.version'],

            'image_id.exact_version': vms['properties.storageProfile'\
                '.imageReference.exactVersion'],

            'zone_id': ['SAZ' for i in range(len(vms))],

            'region_id': vms['location'],

            'host_id': ['Azure' for i in range(len(vms))],

            'resource_group_id': vms['id'].apply(lambda x: x.split('/')[4]),

            'device_type_id': ['Virtual Machine' for i in range(len(vms))],

            'tags': tags,

            'project_id': project_id,

            'environment_id': [
                'Prod' if subscription_id_name[i] == 'ABI BR Prod'
                else 'Non-Prod' if subscription_id_name[i] == 'ABI BR Non-Prod'
                else f'{subscription_id_name[i]} - {project_id[i]}'
                for i in range(len(vms))
            ],

            'azure_status_id': [
                item['properties']['instanceView']\
                    ['statuses'][1]['displayStatus']
                    for item in data['status']
            ],

            'instance_type_id.name': it_info['name'].values,

            'instance_type_id.core_count':  it_info['numberOfCores'].values,

            'instance_type_id.os_disk_size_mb': \
                it_info['osDiskSizeInMB'].values,

            'instance_type_id.resource_disk_size_mb': \
                it_info['resourceDiskSizeInMB'].values,

            'instance_type_id.memory_mb': \
                it_info['memoryInMB'].values,

            'instance_type_id.max_data_disk_count': \
                it_info['maxDataDiskCount'].values,

            'scan_date': [date for i in range(len(vms))]
        }).fillna('')
        return vmTable

    def filter_vms(data):
        new_data = []
        for item in data:
            resource_group = item["id"].split("/")[4].lower()
            computer_name = item["name"].lower()
            conditions = [
                "databricks-rg" in resource_group 
                or "mc_aks-" in resource_group,
                "aks_nodepool" in computer_name 
                or "rtr-" in computer_name,
            ]
            condition = True in conditions
            if not condition:
                new_data.append(item)
        return new_data

    def format(data):
        data['virtual_machine'] = virtual_machine_table_comprehensive\
            .filter_vms(data['virtual_machine'])
        data['status'] = virtual_machine_table_comprehensive\
            .filter_vms(data['status'])
        return virtual_machine_table_comprehensive.create_table(data)



import sqlalchemy

def create_url(db_name, db_host, db_user, db_pass, in_cloud):
#   Equivalent URL:
#   mysql+pymysql://<db_user>:<db_pass>@/<db_name>?unix_socket=<socket_path>/<cloud_sql_instance_name>
    if in_cloud:
        db_socket_dir = "/cloudsql"
        instance_connection_name = os.getenv("INSTANCE_CONNECTION_NAME")
        return sqlalchemy.engine.url.URL.create(
            drivername="mysql+pymysql",
            username=db_user,  # e.g. "my-database-user"
            password=db_pass,  # e.g. "my-database-password"
            database=db_name,  # e.g. "my-database-name"
            host=db_host,
            query={
                "unix_socket": "{}/{}".format(
                db_socket_dir,  # e.g. "/cloudsql"
                instance_connection_name)  # i.e "<PROJECT-NAME>:<INSTANCE-REGION>:<INSTANCE-NAME>"
                }
        )
    else:
        return sqlalchemy.engine.url.URL.create(
            drivername="mysql+pymysql",
            username=db_user,  # e.g. "my-database-user"
            password=db_pass,  # e.g. "my-database-password"
            database=db_name,  # e.g. "my-database-name"
            host=db_host,
        )


# criada a partir do documento: https://cloud.google.com/sql/docs/mysql/connect-functions#configure
def get_engine(db_name, db_host, db_user, db_pass, in_cloud):
    url = create_url(db_name, db_host, db_user, db_pass, in_cloud)

    engine = sqlalchemy.create_engine(
        url
    )
    return engine



from .apis.AzureAPI import AzureAPI
from .apis.GraphAPI import GraphAPI
from .apis.SalesForceAPI import SalesForceAPI
from .fourmatters import virtual_network, virtual_machine
from .logger.Logger import Logger
from .http.HttpListener import HttpListener
from . import gcp

def format(data, table_type):
    formatter_types = {
        'virtual_network_table_std': virtual_network.virtual_network_table_std,
        'virtual_machine_table_comprehensive': virtual_machine.virtual_machine_table_comprehensive 
    }
    return formatter_types.get(table_type).format(data)
    
    



import pandas

class virtual_network_table_std:
    def split_id(df):
        subscriptions = []
        resource_groups = []
        for row in df['id']:
            splited = row.split('/')
            subscriptions.append(splited[2].lower())
            resource_groups.append(splited[4].lower())

        df['subscription_id'] = subscriptions
        df['resource_group_id'] = resource_groups        
        return df

    def expand_properties(df):
        properties = pandas.json_normalize(df['properties'].values)
        df = df.drop('properties', axis=1)
        df = pandas.concat([df, properties], axis=1)
        return df

    def rename_properties(subdf):
        subdf.columns = subdf.columns.str.replace('properties.', '')
        return subdf

    def extract_subnets(df):
        subnets_arr = []
        for val in df['subnets'].values:
            subnets_arr += val

        subdf = pandas.json_normalize(subnets_arr)
        subdf = virtual_network_table_std.rename_properties(subdf)
        df.loc[:,'subnets'] = df.subnets.apply(lambda x: [i['id'] for i in x])
        return df, subdf[subdf.columns[0:9]]

    def extract_peers(df):
        df.loc[:, 'virtualNetworkPeerings'] = \
            df['virtualNetworkPeerings'].apply(lambda x: [i['id'] for i in x])
        return df

    def stringify(df, subdf):
        # stringifying df
        array_columns = [
            'subnets', 
            'virtualNetworkPeerings',
            'addressSpace.addressPrefixes',
            'dhcpOptions.dnsServers'
        ]
        for column in array_columns:
            df[column] = df[column].apply(
                lambda x: ';'.join(x) if type(x) == list else x)

        df.loc[:,'tags'] = df['tags'].apply(lambda x: str(x))

        # stringifying subdf
        subdf['ipConfigurations'] = subdf['ipConfigurations'].apply(
            lambda x: ';'.join([i['id'] for i in x]) if type(x) == list else x)

        subdf['serviceEndpoints'] = subdf['serviceEndpoints'].apply(
            lambda x: ';'.join([i['service'] for i in x]) if \
                type(x) == list else x)
        return df, subdf

    def rm_columns(df, subdf):
        df.pop('id')
        df.pop('etag')
        df.pop('type')
        df.pop('resourceGuid') 
        subdf.pop('id')
        subdf.pop('etag')
        subdf.pop('type')
        return df, subdf

    def format(data):
        df = pandas.json_normalize(data, max_level=0)
        df = virtual_network_table_std.split_id(df)
        df = virtual_network_table_std.expand_properties(df)
        df = virtual_network_table_std.extract_peers(df)
        df, subdf = virtual_network_table_std.extract_subnets(df)
        df, subdf = virtual_network_table_std.stringify(df, subdf)
        df['scan_date'] = pandas.Timestamp.today().strftime("%m-%d-%Y")
        df, subdf = virtual_network_table_std.rm_columns(df, subdf)
        df = df.fillna('')
        subdf = subdf.fillna('')
        return df, subdf




import os, mattlib

def unique_dict(array, key):
    size = len(array)
    unique_key = []
    unique = []
    for i in range(size):
        if array[i][key] not in unique_key:
            unique_key.append(array[i][key])
            unique.append(array[i])
    return unique

api = mattlib.AzureAPI(app_ID=os.getenv('AMBEV_AZURE_APP_ID'),
                        tenant_ID=os.getenv('AMBEV_AZURE_TENANT_ID'),
                        secret_key=os.getenv('AMBEV_AZURE_SECRET_KEY'))

_locations = api.list_locations()
_locations = [
	{
		"subscription": _loc['id'].split('/')[2],
		"location": _loc['id'].split('/')[-1]
	}
	for _loc in _locations
]
locations = unique_dict(_locations, 'location')
_instance_types = api.list_virtual_machine_sizes(locations=locations)



import requests
import os
import json
import pathlib

class BaseMicrosoftAPI:
    def __init__(self, tenant_ID, app_ID, secret_key, scope):
        self.tenant_ID = tenant_ID.rstrip()
        self.app_ID = app_ID.rstrip()
        self.secret_key = secret_key.rstrip()
        self.scope = scope
        self.headers = self.get_auth()

    def get_auth(self):
        token_url = f'https://login.microsoftonline.com/'\
                    f'{self.tenant_ID}/oauth2/v2.0/token'
        auth = {
            'grant_type': 'client_credentials',
            'client_id': self.app_ID,
            'client_secret': self.secret_key,
            'scope': self.scope,
        }
        response = requests.post(token_url, data=auth)
        token = response.json().get('access_token')
        if token != None:
            headers = {'Authorization': f'Bearer {token}'}
            return headers
        else:
            raise Exception(f' BaseMicrosoftAPI authentication failed.\n'\
                            f'Response: {response}')

    def call_api_stream(self, url):
        response = requests.get(url, headers=self.headers)
        return response.text

    def call_api(self, url):
        values = []
        while url != None:
            response = requests.get(url, headers=self.headers)
            status = response.status_code
            if status != 200:
#                print('Response error')
#                print(response.text)
                return None
            response = json.loads(response.text)
            values += response['value']
            if 'nextLink' in response.keys():
                url = response['nextLink']
            if '@odata.nextLink' in response.keys():
                url = response['@odata.nextLink']
            else :
                url = None
        return values



from flask import Flask, request

class HttpListener:
    def __init__(self, port=8080):
        self.port = port
        self.app = Flask('HttpListener')

    def get_data(self, request):
        return request.get_json()

    def respond_get(self):
        try:
            callback_data = callback(*args)
            if callback_data:
                return callback_data, 204
            else:
                return '', 204
        except Exception as e:
            print(e)
            return 'Bad request', 400

    def respond_post(self, request):
        post_data = request.get_json()
        try:
            callback_data = callback(post_data, *args)
            if callback_data:
                return callback_data, 204
            else:
                return '', 204
        except Exception as e:
            print(e)
            return 'Bad request', 400

    def listen(self, callback, args=()):
        @self.app.route('/', methods=['GET', 'POST'])
        def index():
            if request.method == 'GET':
                self.respond_get()
            if request.method == 'POST':
                self.respond_post()

        self.app.run(debug=True,
                     host='0.0.0.0',
                     port=self.port,
                     use_reloader=False)
        



from .BaseMicrosoftAPI import BaseMicrosoftAPI
import requests

class GraphAPI(BaseMicrosoftAPI):
    def __init__(self, tenant_ID, app_ID, secret_key):
       super().__init__(tenant_ID, app_ID, secret_key,
                        'https://graph.microsoft.com/.default')

    def list_users(self, properties=None):
        if properties:
            parameters = ','.join(properties)
            url = f'https://graph.microsoft.com/v1.0/users?$top=999&$select={parameters}'
        else:
            url = 'https://graph.microsoft.com/v1.0/users'
        response = self.call_api(url)
        return response

    def list_subscribed_skus(self):
        url = 'https://graph.microsoft.com/v1.0/subscribedSkus'
        response = self.call_api(url)
        return response

    def office365_active_user_detail(self, params):
        urls = {
            'period': f"https://graph.microsoft.com/v1.0/reports/"\
                      f"getOffice365ActiveUserDetail(period='D{params[1]}')",
            'date': f"http://graph.microsoft.com/v1.0/reports/"\
                    f"getOffice365ActiveUserDetail(date={params[1]})"
        }

        url = urls.get(params[0])
        response = self.call_api_stream(url)
        return response

    def office365_mailbox_usage_detail(self, period):
        supported_periods = [7, 30, 90, 180]
        if period not in supported_periods:
            print(f'office_365_mailbox_usage_detail: please use one of'\
                  f'supported periods: {period}')
            # raise error
            return None
        else:
            url = f"https://graph.microsoft.com/v1.0/reports/"\
                  f"getMailboxUsageDetail(period='D{period}')"
            response = self.call_api_stream(url)
            return response

# Possibilities --------------------------------------------------------------
# users/delta
#    def office365_services_user_counts(self):
# 



from .BaseMicrosoftAPI import BaseMicrosoftAPI

class AzureAPI(BaseMicrosoftAPI):
    def __init__(self, tenant_ID, app_ID, secret_key):
       super().__init__(tenant_ID, app_ID, secret_key,
                        'https://management.core.windows.net/.default')
       self.subscriptions = None
       self.resource_groups = None
       self.servers = None

    def list_subscriptions(self):
        url = 'https://management.azure.com/'\
              'subscriptions?api-version=2020-01-01'
        response = self.call_api(url)
        self.subscriptions = [item['subscriptionId'] for item in response]
        return response

    def subscriptionIds(self):
        subscriptions = self.list_subscriptions()
        return [item['subscriptionId'] for item in subscriptions]

    def list_virtual_networks(self, subscriptions=None):
        if subscriptions is None:
            if self.subscriptions is None:
                self.subscriptions = self.subscriptionIds()
            subscriptions = self.subscriptions
#            subscriptions = self.subscriptionIds()

        virtual_networks = []
        for subscription in subscriptions:
            url = f'https://management.azure.com/subscriptions/{subscription}/'\
                  f'providers/Microsoft.Network/'\
                  f'virtualNetworks?api-version=2021-03-01'
            virtual_networks += self.call_api(url)
        return virtual_networks

    def list_resource_groups(self, subscriptions=None):
        if subscriptions is None:
            if self.subscriptions is None:
                self.subscriptions = self.subscriptionIds()
            subscriptions = self.subscriptions
#            subscriptions = self.subscriptionIds()
        resource_groups = []
        for subscription in subscriptions:
            url = f'https://management.azure.com/subscriptions/{subscription}/'\
                  f'resourcegroups?api-version=2021-04-01'
            resource_groups += self.call_api(url)
        return resource_groups

    def resource_groups_id(self,subscriptions):
        resource_groups = self.list_resource_groups(subscriptions)
        return [item['id'] for item in resource_groups]

    def list_virtual_machines(self, subscriptions=None, status=False):
        if subscriptions is None:
            if self.subscriptions is None:
                self.subscriptions = self.subscriptionIds()
            subscriptions = self.subscriptions
#            subscriptions = self.subscriptionIds()

        virtual_machines = []
        for subscription in subscriptions:
            url = f'https://management.azure.com/subscriptions/{subscription}/'\
                  f'providers/Microsoft.Compute/'\
                  f'virtualMachines?api-version=2021-03-01&statusOnly={status}'
            virtual_machines += self.call_api(url)
        return virtual_machines



    def list_public_ip_addresses(self, subscriptions=None):
        if subscriptions is None:
            if self.subscriptions is None:
                self.subscriptions = self.subscriptionIds()
            subscriptions = self.subscriptions
#            subscriptions = self.subscriptionIds()

        public_ips = []
        for subscription in subscriptions:
            url = f'https://management.azure.com/subscriptions/{subscription}/'\
                  f'providers/Microsoft.Network/'\
                  f'publicIPAddresses?api-version=2021-02-01'
            public_ips += self.call_api(url)
        return public_ips

    def list_locations(self, subscriptions=None):
        if subscriptions is None:
            if self.subscriptions is None:
                self.subscriptions = self.subscriptionIds()
            subscriptions = self.subscriptions
        
        locations = []
        for subscription in subscriptions:
            url = f'https://management.azure.com/subscriptions/{subscription}/'\
                  f'locations?api-version=2020-01-01'
            locations += self.call_api(url)
        return locations

    # Microsoft states this API is deprecated.
    def list_virtual_machine_sizes(self, subscriptions=None, locations=None):
        if subscriptions is None:
            if self.subscriptions is None:
                self.subscriptions = self.subscriptionIds()
            subscriptions = self.subscriptions

        if locations is None:
            locations = self.list_locations()

        sizes = []
        for location in locations:
            subscription = location['subscription']
            location_name = location['location']
            url = f'https://management.azure.com/subscriptions/'\
                  f'{subscription}/providers/Microsoft.Compute/locations/'\
                  f'{location_name}/vmSizes?api-version=2021-07-01'
            result = self.call_api(url)
            if result:
                sizes += result
        return sizes

    def list_network_interfaces(self, subscriptions=None):
        if subscriptions is None:
            if self.subscriptions is None:
                self.subscriptions = self.subscriptionIds()
            subscriptions = self.subscriptions

        net_interfaces = []
        for subscription in subscriptions:
            url = f'https://management.azure.com/subscriptions/{subscription}'\
                  f'/providers/Microsoft.Network/'\
                  f'networkInterfaces?api-version=2021-03-01'
            net_interfaces += self.call_api(url)
        return net_interfaces



    def list_application_gateways(self, subscriptions=None):
        if subscriptions is None:
            if self.subscriptions is None:
                self.subscriptions = self.subscriptionIds()
            subscriptions = self.subscriptions

        app_gateways = []
        for subscription in subscriptions:
            url = f'https://management.azure.com/subscriptions/{subscription}'\
                  f'/providers/Microsoft.Network/'\
                  f'applicationGateways?api-version=2021-05-01'
            app_gateways += self.call_api(url)
        return app_gateways

    def list_databricks_workspaces(self, subscriptions=None):
        if subscriptions is None:
            if self.subscriptions is None:
                self.subscriptions = self.subscriptionIds()
            subscriptions = self.subscriptions

        databricks_workspaces = []
        for subscription in subscriptions:
            url = f'https://management.azure.com/subscriptions/{subscription}'\
                  f'/providers/Microsoft.Databricks/'\
                  f'workspaces?api-version=2018-04-01'
            databricks_workspaces += self.call_api(url)
        return databricks_workspaces

    def list_factories(self, subscriptions=None):
        if subscriptions is None:
            if self.subscriptions is None:
                self.subscriptions = self.subscriptionIds()
            subscriptions = self.subscriptions

        factories = []
        for subscription in subscriptions:
            url = f'https://management.azure.com/subscriptions/{subscription}'\
                  f'/providers/Microsoft.DataFactory/'\
                  f'factories?api-version=2018-06-01'
            factories += self.call_api(url)
        return factories

    def list_mssql_server(self, resource_groups=None):

        if self.subscriptions is None:
            self.subscriptions = self.subscriptionIds()
        subscriptions = self.subscriptions

        if resource_groups is None:
            if self.resource_groups is None:
                self.resource_groups = self.resource_groups_id(subscriptions)
            resource_groups = self.resource_groups

        subs = []
        rgs = []
        for row in resource_groups:
            splited = row.split('/')
            subs.append(splited[2].lower())
            rgs.append(splited[4].lower())

        servers = []
        for i in range(len(subs)):
                url = f'https://management.azure.com/subscriptions/{subs[i]}'\
                    f'/resourceGroups/{rgs[i]}/providers/Microsoft.Sql/'\
                    f'servers?api-version=2021-11-01-preview'
                servers += self.call_api(url)
        return servers

    def servers_ids(self,resource_groups):
        servers = self.list_mssql_server(resource_groups)
        return [item['id'] for item in servers]

    def list_mssql_database(self, servers=None):

        if self.subscriptions is None:
            self.subscriptions = self.subscriptionIds()
        subscriptions = self.subscriptions

        if self.resource_groups is None:
            self.resource_groups = self.resource_groups_id(subscriptions)
        resource_groups = self.resource_groups

        if servers is None:
            if self.servers is None:
                self.servers = self.servers_ids(resource_groups)
            servers = self.servers        

        subs = []
        rgs = []
        server_names = []
        for row in servers:
            splited = row.split('/')
            subs.append(splited[2].lower())
            rgs.append(splited[4].lower())
            server_names.append(splited[8].lower())

        databases = []
        for i in range(len(server_names)):
                    url = f'https://management.azure.com/subscriptions/{subs[i]}'\
                        f'/resourceGroups/{rgs[i]}/providers/Microsoft.Sql/servers/{server_names[i]}/'\
                        f'databases?api-version=2021-11-01-preview'
                    databases += self.call_api(url)
        return databases

    def list_resources(self, subscriptions=None):
        if subscriptions is None:
            if self.subscriptions is None:
                self.subscriptions = self.subscriptionIds()
            subscriptions = self.subscriptions

        resources = []
        for subscription in subscriptions:
            url = f'https://management.azure.com/subscriptions/{subscription}/'\
                  f'resources?api-version=2021-04-01'
            resources += self.call_api(url)
        return resources



import requests
import os
import json

# Authorization must contain fields:
#    type
#    domain
#    consumer-key
#    consumer-secret
#    username
#    password

class SalesForceAPI:
    def __init__(self, authorization):
        self.url = f'https://{authorization["domain"]}.my.salesforce.com'
        if authorization['type'] == 'username-password':
            self.headers = self.__get_auth_user(authorization)
        if authorization['type'] == 'web server':
            self.headers = self.__get_auth_web_server(authorization)
        # headers must be: 
        # { 'Authorization': <token>, 'X-PrettyPrint': 1 }

    def __get_auth_user(self, authorization):
        domain = authorization['domain']
        auth = {
            'grant_type': 'password',
            'client_id': authorization['consumer-key'].rstrip(),
            'client_secret': authorization['consumer-secret'].rstrip(),
            'username': authorization['username'].rstrip(),
            'password': authorization['password'].rstrip()
        }
        url = f'{self.url}/services/oauth2/token'
        response = requests.post(url, data=auth)
        token = response.json().get('access_token')
        if token != None:
            headers = {'Authorization': f'Bearer {token}'}
            return headers
        else:
            raise Exception(f'SalesForceAPI authentication failed.\n'\
                            f'Response: {response}')

    def __get_auth_web_server(self, authorization):
        pass

    def user(self, fields=None):
        if not fields:
            fields = [
                'Id', 'FirstName', 
                'LastName', 'Username', 'Email',
                'IsActive', 'LastLoginDate',
                'UserType', 'ProfileId'
            ]
        fields = ',+'.join(fields)
        query = f'SELECT+{fields}+from+User'
        request_url = f'{self.url}/services/data/v53.0/query?q={query}'
        response = self.call_api(request_url)
        return response

    def profile(self, fields=None):
        if not fields:
            fields = [
                'Id', 'Name', 'UserLicenseId'
            ]
        fields = ',+'.join(fields)
        query = f'SELECT+{fields}+from+Profile'
        request_url = f'{self.url}/services/data/v53.0/query?q={query}'
        response = self.call_api(request_url)
        return response

    def PermissionSetLicense(self, fields=None):
        if not fields:
            fields = [
                'Id', 'PermissionSetLicenseKey', 'DeveloperName', 'TotalLicenses',
                'UsedLicenses', 'Status'
            ]
        fields = ',+'.join(fields)
        query = f'SELECT+{fields}+from+PermissionSetLicense'
        request_url = f'{self.url}/services/data/v53.0/query?q={query}'
        response = self.call_api(request_url)
        return response

    def TenantUsageEntitlement(self, fields=None):
        if not fields:
            fields = [
                'MasterLabel', 'AmountUsed', 'CurrentAmountAllowed', 'Frequency', 
                'IsPersistentResource', 'UsageDate' 
            ]
        fields = ',+'.join(fields)
        query = f'SELECT+{fields}+from+TenantUsageEntitlement'
        request_url = f'{self.url}/services/data/v53.0/query?q={query}'
        response = self.call_api(request_url)
        return response

    def ActiveFeatureLicenseMetric(self, fields=None):
        if not fields:
            fields = [
                'ActiveUserCount', 'AssignedUserCount', 'FeatureType',
                'MetricsDate', 'TotalLicenseCount'
            ]
        fields = ',+'.join(fields)
        query = f'SELECT+{fields}+from+ActiveFeatureLicenseMetric'
        request_url = f'{self.url}/services/data/v53.0/query?q={query}'
        response = self.call_api(request_url)
        return response
    
    def userLicense(self, fields=None):
        if not fields:
            fields = [
                'Id', 'LicenseDefinitionKey', 'Name',
                'TotalLicenses', 'UsedLicenses', 'Status', 
            ]
        fields = ',+'.join(fields)
        query = f'SELECT+{fields}+from+UserLicense'
        request_url = f'{self.url}/services/data/v53.0/query?q={query}'
        response = self.call_api(request_url)
        return response
           

    def call_api(self, url):
        response = requests.get(url, headers=self.headers)
        response = json.loads(response.text)
        return response



import os
import pandas as pd
import mattlib

auth = {
    'type': 'username-password',
    'domain': os.getenv('SALESFORCE_DOMAIN'),
    'consumer-key': os.getenv('SALESFORCE_CONSUMER_KEY'),
    'consumer-secret': os.getenv('SALESFORCE_CONSUMER_SECRET'),
    'username': os.getenv('SALESFORCE_USERNAME'),
    'password': os.getenv('SALESFORCE_PASSWORD')
}
api = mattlib.SalesForceAPI(auth)
users = api.user()
udf = pd.json_normalize(users.json()['records'])

profiles = api.profile()
pdf = pd.json_normalize(profiles.json()['records'])

userLicenses = api.userLicense()
uldf = pd.json_normalize(userLicenses.json()['records'])



from datetime import datetime

class Logger:
    def __init__(self, logfile='log.txt', interrupt=True):
        if type(logfile) == str:
            self.file_pointer = False
            self.file = False
            self.file_name = logfile
        else:
            self.file_pointer = True
            self.file = logfile
            self.file_name = False

        self.interrupt = interrupt

    def log(self, msg):
        time = datetime.now().strftime('%d-%m-%Y  %H:%M:%S:%f')
        log_msg = f'{time}\t{msg}\n'
        if self.file_pointer:
            print(log_msg.strip(), file=self.file)  
        else:
            with open(self.file_name, 'a') as outfile:
                outfile.write(log_msg)
        
    def run(self, function, parameters=(), secret=False):
        if not secret:
            self.log(f'executing {function} with parameters: {parameters}')
        else:
            self.log(f'executing {function}')
        try:
            if type(parameters) == tuple:
                result = function(*parameters)
            if type(parameters) == dict:
                result = function(**parameters)
            self.log(f'concluded {function}')
            return result
        except Exception as e:
            self.log(f'Error: {e}')
            if self.interrupt:
                exit()
            else:
                return None




import base64

def get_message(envelope, decode=True):
    if envelope == None:
        raise Exception("no Pub/Sub message received")

    try:
        pubsub_message = envelope["message"]
        if decode:
            message = base64.b64decode(pubsub_message["data"])\
                .decode("utf-8")\
                .strip()
        else:
            message = pubsub_message['data']
        return message
    except:
        raise Exception("invalid Pub/Sub message format")
